"""
Module containing Links.

These objects are used in service layer responses to represent,
abstractly, a link to another entity. The HTTP layer then converts these
link objects to concrete URLs.
"""

from transiter import models


class Link:
    kwargs = {}

    def __eq__(self, other):
        return type(self) == type(other) and self.kwargs == other.kwargs

    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            ", ".join("{}={}".format(*item) for item in self.kwargs.items()),
        )


class InternalDocumentationLink(Link):
    pass


class SystemsIndexLink(Link):
    pass


class FeedEntityLink(Link):
    def __init__(self, feed: models.Feed):
        self.kwargs = {"system_id": feed.system.id, "feed_id": feed.id}


class FeedEntityUpdatesLink(Link):
    def __init__(self, feed: models.Feed):
        self.kwargs = {"system_id": feed.system.id, "feed_id": feed.id}


class FeedsInSystemIndexLink(Link):
    def __init__(self, system: models.System):
        self.kwargs = {"system_id": system.id}


class StopEntityLink(Link):
    def __init__(self, stop: models.Stop):
        self.kwargs = {"system_id": stop.system.id, "stop_id": stop.id}


class StopsInSystemIndexLink(Link):
    def __init__(self, system: models.System):
        self.kwargs = {"system_id": system.id}


class SystemEntityLink(Link):
    def __init__(self, system: models.System):
        self.kwargs = {"system_id": system.id}


class RouteEntityLink(Link):
    def __init__(self, route: models.Route):
        self.kwargs = {"system_id": route.system.id, "route_id": route.id}


class RoutesInSystemIndexLink(Link):
    def __init__(self, system: models.System):
        self.kwargs = {"system_id": system.id}


class TripEntityLink(Link):
    def __init__(self, trip: models.Trip):
        self.kwargs = {
            "system_id": trip.route.system.id,
            "route_id": trip.route.id,
            "trip_id": trip.id,
        }



"""
The route service is used to retrieve data about routes.
"""

import enum

from transiter import exceptions
from transiter.data import dbconnection
from transiter.data.dams import routedam, systemdam, servicemapdam
from transiter.models import Alert
from transiter.services import links, constants as c


@dbconnection.unit_of_work
def list_all_in_system(system_id, return_links=True):
    """
    Get data on all routes in a system.

    This function returns a list of dictionaries, one dictionary for each
    route. The dictionary for a specific route contains
     * the route's short representation,
     * its status under key 'status',
     * and optionally a link to the route.

    :param system_id: the ID of the system
    :type system_id: str
    :param return_links: whether to return links
    :type return_links: bool
    :return: the list described above
    :rtype: list
    """
    system = systemdam.get_by_id(system_id, only_return_active=True)
    if system is None:
        raise exceptions.IdNotFoundError
    response = []
    routes = list(routedam.list_all_in_system(system_id))
    route_pk_to_status = _construct_route_pk_to_status_map(route.pk for route in routes)
    for route in routes:
        route_response = {**route.to_dict(), c.STATUS: route_pk_to_status[route.pk]}
        if return_links:
            route_response[c.HREF] = links.RouteEntityLink(route)
        response.append(route_response)
    return response


@dbconnection.unit_of_work
def get_in_system_by_id(system_id, route_id, return_links=True):
    """
    Get data for a specific route in a specific system.

    This function returns a dictionary containing,
     * the route's long representation,
     * its status under key 'status',
     * its periodicity under key 'periodicity',
     * its alerts under key 'alerts',
     * all of its service maps for whose group use_for_stops_in_route is true,
       under key 'service_maps',
     * and optionally a link to the route.

    :param system_id: the system ID
    :type system_id: str
    :param route_id: the route ID
    :type route_id: str
    :param return_links: whether to return links
    :type return_links: bool
    :return: the dictionary described above
    :rtype: dict
    """
    route = routedam.get_in_system_by_id(system_id, route_id)
    if route is None:
        raise exceptions.IdNotFoundError
    status = _construct_route_status(route.pk)
    periodicity = routedam.calculate_periodicity(route.pk)
    if periodicity is not None:
        periodicity = int(periodicity / 6) / 10
    response = {
        **route.to_large_dict(),
        c.PERIODICITY: periodicity,
        c.STATUS: status,
        c.ALERTS: [alert.to_large_dict() for alert in route.route_statuses],
        c.SERVICE_MAPS: [],
    }

    for group, service_map in servicemapdam.list_groups_and_maps_for_stops_in_route(
        route.pk
    ):
        service_map_response = {c.GROUP_ID: group.id, c.STOPS: []}
        if service_map is not None:
            for entry in service_map.vertices:
                stop_response = entry.stop.to_dict()
                if return_links:
                    stop_response[c.HREF] = links.StopEntityLink(entry.stop)
                service_map_response[c.STOPS].append(stop_response)
        response[c.SERVICE_MAPS].append(service_map_response)

    return response


class Status(str, enum.Enum):
    """Enum containing the possible statuses for a route."""

    NO_SERVICE = "NO_SERVICE"
    GOOD_SERVICE = "GOOD_SERVICE"
    PLANNED_SERVICE_CHANGE = "PLANNED_SERVICE_CHANGE"
    UNPLANNED_SERVICE_CHANGE = "UNPLANNED_SERVICE_CHANGE"
    DELAYS = "DELAYS"


def _construct_route_status(route_pk):
    """
    Construct the status for a specific route.

    See _construct_route_pk_to_status_map for documentation on how this is
    constructed.

    :param route_pk: the route's PK
    :type route_pk: int
    :return: the route's status
    :rtype: Status
    """
    return _construct_route_pk_to_status_map([route_pk])[route_pk]


def _construct_route_pk_to_status_map(route_pks_iter):
    """
    Construct the statuses for multiple routes.

    The algorithm constructs the status by first examining the highest priority
    alerts for a specific route. If there are alerts then,
     * If one of these alerts has the effect SIGNIFICANT_DELAYs, then the status
       of the route is DELAYS.
     * Otherwise, if one of the alerts has planned false, then the status is
       UNPLANNED_SERVICE_CHANGE.
     * Otherwise, the status is PLANNED_SERVICE_CHANGE.

    If there are now no alerts then it is checked to see if there are active
    trips for this route.
     * If so, the status is GOOD_SERVICE.
     * Otherwise, the status is NO_SERVICE.

    :param route_pks_iter: iterator of the routes's PKs
    :type route_pks_iter: iter
    :return: list of TripStatus objects
    :rtype: list
    """
    route_pks = {route_pk for route_pk in route_pks_iter}

    route_pk_to_alerts = routedam.get_route_pk_to_highest_priority_alerts_map(route_pks)

    route_pk_to_status = {route_pk: Status.NO_SERVICE for route_pk in route_pks}
    for route_pk, alerts in route_pk_to_alerts.items():
        if len(alerts) == 0:
            continue
        causes = set(alert.cause for alert in alerts)
        effects = set(alert.effect for alert in alerts)
        if Alert.Effect.SIGNIFICANT_DELAYS in effects:
            route_pk_to_status[route_pk] = Status.DELAYS
        elif Alert.Cause.ACCIDENT in causes:
            route_pk_to_status[route_pk] = Status.UNPLANNED_SERVICE_CHANGE
        else:
            route_pk_to_status[route_pk] = Status.PLANNED_SERVICE_CHANGE
        route_pks.remove(route_pk)

    for route_pk in routedam.list_route_pks_with_current_service(route_pks):
        route_pk_to_status[route_pk] = Status.GOOD_SERVICE

    return route_pk_to_status



from sqlalchemy import inspect
from sqlalchemy.ext.declarative import declarative_base


class _BaseModel:
    """
    Basic model that all Transiter models inherit from.

    This class defines an equals method for use in testing, and representation
    functions for outputting dictionary representations of entities in the service
    layer.
    """

    # The equals method is designed for testing purposes only.
    def __eq__(self, other):
        for column in inspect(type(self)).columns.keys():
            if getattr(self, column) != getattr(other, column):
                # print(
                #    'Values for attribute "{}" don\'t match: "{}" != "{}"'.format(
                #        column, str(getattr(self, column)), str(getattr(other, column))
                #    )
                # )
                return False
        return True

    # The __repr__ method is designed for testing purposes only.
    def __repr__(self):
        attributes = []
        for column in inspect(type(self)).columns.keys():
            attributes.append("{}={}".format(column, getattr(self, column)))
        return "{}({})".format(type(self).__name__, ", ".join(attributes))

    # This is a hack to get around a problem in SQL Alchemy where the hash is evaluated
    # It should not be relied on and will be replaced at some point by an exception!
    def __hash__(self):
        return id(self)


Base = declarative_base(cls=_BaseModel)


class ToDictMixin:

    __dict_columns__ = None
    __large_dict_columns__ = None

    def to_dict(self) -> dict:
        return self._to_dict(self.__dict_columns__)

    def to_large_dict(self) -> dict:
        return self._to_dict(self.__large_dict_columns__)

    def _to_dict(self, columns) -> dict:
        if columns is None:
            raise NotImplementedError
        return {column.key: getattr(self, column.key, None) for column in columns}



ALERTS = "alerts"
CHILD_STOPS = "child_stops"
COUNT = "count"
DIRECTION = "direction"
DIRECTIONS = "directions"
ERROR = "error"
FEEDS = "feeds"
GROUP_ID = "group_id"
HREF = "href"
LAST_STOP = "last_stop"
PARENT_STOP = "parent_stop"
PERIODICITY = "periodicity"
ROUTE = "route"
ROUTES = "routes"
SERVICE_MAPS = "service_maps"
STATUS = "status"
STOP = "stop"
STOP_ID = "stop_id"
STOP_TIMES = "stop_times"
STOPS = "stops"
TRIP = "trip"
UPDATES = "updates"



"""
The GTFS Realtime Util contains the logic for reading feeds of this format.

The official reference is here: https://gtfs.org/reference/realtime/v2/
"""
import datetime

import pytz
from google.transit import gtfs_realtime_pb2

from transiter import models


def create_parser(gtfs_realtime_pb2_module=None, post_pb2_parsing_function=None):
    """
    Create a GTFS Realtime parser.

    The returned parser performs the following steps in order:

    (1) Read the GTFS Realtime binary data into a JSON structure using either
        the default GTFS Realtime protobuf modules or a custom module.
        The custom is for when a GTFS Realtime extension is being used.

    (2) Optionally apply a post-pb2-parsing function to the JSON structure.
        This is generally use for merging GTFS Realtime extension data into
        the main JSON structure. Clean-up of the data within the feed is
        easier with a trip cleaner in (4).

    (3) Convert the JSON structure into various Transiter Python objects.
        These objects are SQL Alchemy models that correspond to database tables.

    :param gtfs_realtime_pb2_module: optional, the protobuffers module to use
        to read the feed. Defaults to the official module on PyPI.
    :param post_pb2_parsing_function: the function described in (2). It takes
        a single argument, which is the JSON data structure. It returns nothing;
        rather, it alterts the JSON structure in place.
    :return: the parser
    """
    # Additional arguments are accepted for forwards compatibility
    # noinspection PyUnusedLocal
    def parser(binary_content, *args, **kwargs):
        gtfs_data = read_gtfs_realtime(binary_content, gtfs_realtime_pb2_module)
        # print(json.dumps(gtfs_data))
        if post_pb2_parsing_function is not None:
            post_pb2_parsing_function(gtfs_data)
        (feed_time, trips) = transform_to_transiter_structure(
            gtfs_data, "America/New_York"
        )
        return trips

    return parser


built_in_parser = create_parser()


def read_gtfs_realtime(content, gtfs_realtime_pb2_module=None):
    """
    Convert a binary GTFS Realtime feed to a JSON-like data structure

    :param content: GTFS realtime binary data
    :param gtfs_realtime_pb2_module: GTFS realtime module
    :return: the data in a JSON-like data structure
    """
    if gtfs_realtime_pb2_module is None:
        gtfs_realtime_pb2_module = gtfs_realtime_pb2

    gtfs_feed = gtfs_realtime_pb2_module.FeedMessage()
    gtfs_feed.ParseFromString(content)
    return _read_protobuf_message(gtfs_feed)


def _read_protobuf_message(message):
    """
    Convert a protobuf message into a JSON-like structure.

    This is not an exhaustive converter but mean to be sufficient for GTFS
    realtime

    :param message: a google.protobuf.message.Message object
    :return: a dictionary of {key: value} for fields in the Message. The
    function is recursive so if value is also a protobuf then that will be
    expanded
    """
    d = {}
    for (descriptor, value) in message.ListFields():
        # Iterate through the possible protobuf field types to decide how to
        # parse this value
        if descriptor.type == descriptor.TYPE_MESSAGE:
            parsing_function = _read_protobuf_message
        elif descriptor.type == descriptor.TYPE_ENUM:

            def parsing_function(value):
                return descriptor.enum_type.values_by_number[value].name

        else:

            def parsing_function(value):
                return value

        # Then parse it
        if descriptor.label == descriptor.LABEL_REPEATED:
            parsed_value = [parsing_function(v) for v in value]
        else:
            parsed_value = parsing_function(value)

        d[descriptor.name] = parsed_value
    return d


def transform_to_transiter_structure(data, timezone_str=None):
    """
    Transform GTFS Realtime data in JSON like format to Transiter models

    :param data: the data
    :param timezone_str: a string decribing the timezone
    :return: feed_time, trips, route_ids in the feed
    """
    transformer = _GtfsRealtimeToTransiterTransformer(data, timezone_str)
    return transformer.transform()


ACTIVE_PERIOD = "active_period"
AGENCY_ID = "agency_id"
ALERT = "alert"
CAUSE = "cause"
DESCRIPTION_TEXT = "description_text"
EFFECT = "effect"
END = "end"
ENTITY = "entity"
HEADER_TEXT = "header_text"
ID = "id"
INFORMED_ENTITY = "informed_entity"
ROUTE_ID = "route_id"
START = "start"
TRANSLATION = "translation"
TEXT = "text"
URL = "url"


class _GtfsRealtimeToTransiterTransformer:
    def __init__(self, raw_data, timezone_str=None):
        self._raw_data = raw_data
        self._trip_id_to_raw_entities = {}
        self._trip_id_to_transformed_entity = {}
        self._trip_id_to_trip_model = {}
        self._transformed_metadata = {}
        self._feed_time = None
        self._timestamp_to_datetime_cache = {}
        if timezone_str is None:
            self._timezone = None
        else:
            self._timezone = pytz.timezone(timezone_str)

    def transform(self):
        self._transform_feed_metadata()
        self._group_trip_entities()
        self._transform_trip_base_data()
        self._transform_trip_stop_events()
        self._update_stop_event_indices()
        return (
            self._feed_time,
            list(self._trip_id_to_trip_model.values())
            + self.build_alerts(self._raw_data),
        )

    def build_alerts(self, raw_data):
        def get_text(alert_data_, key):
            texts = alert_data_.get(key, {}).get(TRANSLATION, [])
            if len(texts) == 0:
                return None
            return texts[0][TEXT]

        def get_enum_value(enum, key, default):
            try:
                return enum[key]
            except KeyError:
                return default

        alerts = []
        for entity in raw_data.get(ENTITY, []):
            alert_id = entity.get(ID)
            alert_data = entity.get(ALERT)
            if alert_id is None or alert_data is None:
                continue
            alert = models.Alert(
                id=alert_id,
                cause=get_enum_value(
                    models.Alert.Cause,
                    alert_data.get(CAUSE),
                    models.Alert.Cause.UNKNOWN_CAUSE,
                ),
                effect=get_enum_value(
                    models.Alert.Effect,
                    alert_data.get(EFFECT),
                    models.Alert.Effect.UNKNOWN_EFFECT,
                ),
                header=get_text(alert_data, HEADER_TEXT),
                description=get_text(alert_data, DESCRIPTION_TEXT),
                url=get_text(alert_data, URL),
                start_time=self._timestamp_to_datetime(
                    alert_data.get(ACTIVE_PERIOD, {}).get(START)
                ),
                end_time=self._timestamp_to_datetime(
                    alert_data.get(ACTIVE_PERIOD, {}).get(START)
                ),
            )
            informed_entities = alert_data.get(INFORMED_ENTITY, {})
            alert.agency_ids = [
                informed_entity[AGENCY_ID]
                for informed_entity in informed_entities
                if AGENCY_ID in informed_entity
            ]
            alert.route_ids = [
                informed_entity[ROUTE_ID]
                for informed_entity in informed_entities
                if ROUTE_ID in informed_entity
            ]
            alerts.append(alert)
        return alerts

    def _transform_feed_metadata(self):
        self._feed_time = self._timestamp_to_datetime(
            self._raw_data["header"]["timestamp"]
        )
        self._transformed_metadata = {"timestamp": self._feed_time}

    def _group_trip_entities(self):
        def attach_entity(entity_key, entity):
            trip_descriptor = entity.get("trip", {})
            trip_id = trip_descriptor.get("trip_id", None)
            if trip_id is None:
                return
            self._trip_id_to_raw_entities.setdefault(trip_id, {})
            self._trip_id_to_raw_entities[trip_id]["trip"] = trip_descriptor
            self._trip_id_to_raw_entities[trip_id][entity_key] = entity

        for main_entity in self._raw_data.get("entity", []):
            if "trip_update" in main_entity:
                attach_entity("trip_update", main_entity["trip_update"])
            if "vehicle" in main_entity:
                attach_entity("vehicle", main_entity["vehicle"])
            # if "alert" in main_entity:
            #    self._alerts_raw_data.append(main_entity["alert"])

    def _transform_trip_base_data(self):
        for trip_id, entity in self._trip_id_to_raw_entities.items():
            trip_data = entity.get("trip", {})

            trip = models.TripLight()
            trip.id = trip_id
            trip.route_id = trip_data.get("route_id", None)
            trip.direction_id = trip_data.get("direction_id", None)
            start_date_str = trip_data.get("start_date", None)
            if start_date_str is not None:
                start_dt = datetime.datetime(
                    year=int(start_date_str[0:4]),
                    month=int(start_date_str[4:6]),
                    day=int(start_date_str[6:8]),
                )
                trip.start_time = self._localize_datetime(start_dt, naive=True)

            trip.vehicle_id = (
                entity.get("trip_update", {}).get("vehicle", {}).get("id", None)
            )

            vehicle_data = entity.get("vehicle", {})
            trip.last_update_time = self._timestamp_to_datetime(
                vehicle_data.get("timestamp", None)
            )
            if trip.last_update_time is None:
                trip.last_update_time = self._transformed_metadata.get(
                    "timestamp", None
                )
            raw_current_status = vehicle_data.get("current_status", None)
            if raw_current_status is not None:
                trip.current_status = models.Trip.TripStatus[raw_current_status]
            trip.current_stop_sequence = vehicle_data.get("current_stop_sequence", 0)
            trip.current_stop_id = vehicle_data.get("stop_id", None)
            self._trip_id_to_trip_model[trip_id] = trip

    def _transform_trip_stop_events(self):
        for trip_id, trip in self._trip_id_to_trip_model.items():
            entity = self._trip_id_to_raw_entities[trip_id]

            trip_update = entity.get("trip_update", {})
            stop_time_updates = []

            for stop_time_update_data in trip_update.get("stop_time_update", []):
                stop_time_update = models.TripStopTime.from_feed(
                    trip_id=trip_id,
                    stop_id=stop_time_update_data["stop_id"],
                    arrival_time=self._timestamp_to_datetime(
                        stop_time_update_data.get("arrival", {}).get("time", None)
                    ),
                    arrival_delay=stop_time_update_data.get("arrival", {}).get("delay"),
                    arrival_uncertainty=stop_time_update_data.get("arrival", {}).get(
                        "uncertainty"
                    ),
                    departure_time=self._timestamp_to_datetime(
                        stop_time_update_data.get("departure", {}).get("time", None)
                    ),
                    departure_delay=stop_time_update_data.get("departure", {}).get(
                        "delay"
                    ),
                    departure_uncertainty=stop_time_update_data.get(
                        "departure", {}
                    ).get("uncertainty"),
                    track=stop_time_update_data.get("track", None),
                )
                stop_time_updates.append(stop_time_update)
            trip.stop_times = stop_time_updates

    def _update_stop_event_indices(self):
        for trip_id, trip in self._trip_id_to_trip_model.items():
            index = trip.current_stop_sequence
            for stop_time_update in trip.stop_times:
                stop_time_update.stop_sequence = index
                index += 1

    def _timestamp_to_datetime(self, timestamp):
        if timestamp is None or timestamp == 0:
            return None
        if timestamp not in self._timestamp_to_datetime_cache:
            utc_dt_naive = datetime.datetime.utcfromtimestamp(timestamp)
            utc_dt = pytz.UTC.localize(utc_dt_naive)
            self._timestamp_to_datetime_cache[timestamp] = self._localize_datetime(
                utc_dt
            )
        return self._timestamp_to_datetime_cache[timestamp]

    def _localize_datetime(self, dt, naive=False):
        if self._timezone is None:
            return dt
        if naive:
            return self._timezone.localize(dt)
        else:
            return dt.astimezone(self._timezone)


class TripDataCleaner:
    """
    The TripDataCleaner provides a mechanism for cleaning valid and removing
    invalid Trips and TripStopTimes.

    To use the TripDataCleaner, you must provide a number of cleaning functions.
    Trip cleaning functions accept two arguments - the current FeedUpdate
    and the Trip being cleaned,
    and perform some operations to clean up the trip - for example, switching
    the direction of Trips in a given route to compensate for a known bug in a
    Transit agency's data feed. If the cleaner returns
    False, then that Trip is removed from the collection. TripStopTime cleaners
    work identically.

    After initializing the cleaner with cleaning functions, a list of Trips
    is passed into its clean method. The cleaners operate on all of the Trips
    and their contained TripStopTimes, remove entities based on cleaner function
    results, and returns the list of cleaned trips.
    """

    def __init__(self, trip_cleaners, stop_time_cleaners):
        """
        Initialize a new TripDataCleaner

        :param trip_cleaners: list of Trip cleaning functions
        :param stop_time_cleaners: list of TripStopTime cleaning functions
        """
        self._trip_cleaners = trip_cleaners
        self._stop_time_cleaners = stop_time_cleaners

    def clean(self, trips):
        """
        Clean a collection of trips.

        :param trips: the trips to clean
        :return: the cleaned trips with bad trips removed
        """
        trips_to_keep = []
        for trip in trips:
            if not isinstance(trip, models.Trip):
                trips_to_keep.append(trip)
                continue
            result = True
            for trip_cleaner in self._trip_cleaners:
                result = trip_cleaner(trip)
                if not result:
                    break
            if not result:
                continue

            for stop_time_update in trip.stop_times:
                for stop_time_cleaner in self._stop_time_cleaners:
                    stop_time_cleaner(stop_time_update)

            trips_to_keep.append(trip)

        return trips_to_keep



import enum

from sqlalchemy import (
    Column,
    Enum,
    TIMESTAMP,
    Integer,
    String,
    Boolean,
    ForeignKey,
    UniqueConstraint,
)
from sqlalchemy.orm import relationship

from .base import Base, ToDictMixin
from .updatableentity import updatable_entity


@updatable_entity
class Trip(ToDictMixin, Base):
    __tablename__ = "trip"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    route_pk = Column(Integer, ForeignKey("route.pk"), nullable=False)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    class TripStatus(enum.Enum):
        SCHEDULED = 1
        INCOMING_AT = 2
        STOPPED_AT = 3
        IN_TRANSIT_TO = 4

    direction_id = Column(Boolean)
    start_time = Column(TIMESTAMP(timezone=True))

    last_update_time = Column(TIMESTAMP(timezone=True))
    vehicle_id = Column(String)
    current_status = Column(Enum(TripStatus))
    current_stop_sequence = Column(Integer)

    route_id = None
    stop_id = None

    source = relationship("FeedUpdate", cascade="none")
    route = relationship("Route", back_populates="trips", cascade="none")
    stop_times = relationship(
        "TripStopTime",
        back_populates="trip",
        order_by="TripStopTime.stop_sequence",
        cascade="all, delete-orphan",
        cascade_backrefs=False,
    )

    __table_args__ = (UniqueConstraint(route_pk, id),)

    __dict_columns__ = [id]
    __large_dict_columns__ = [
        id,
        direction_id,
        start_time,
        last_update_time,
        current_status,
        current_stop_sequence,
        vehicle_id,
    ]

    __mapping_columns__ = [pk, route_pk, source_pk] + __large_dict_columns__

    def to_mapping(self):
        return self._to_dict(self.__mapping_columns__)

    @classmethod
    def from_feed(cls, *args, **kwargs):
        return TripLight()


class TripLight:

    route_pk = None

    def to_mapping(self):
        attrs = [
            "pk",
            "route_pk",
            "source_pk",
            "id",
            "direction_id",
            "start_time",
            "last_update_time",
            "current_status",
            "current_stop_sequence",
            "vehicle_id",
        ]
        d = {}
        for attr in attrs:
            d[attr] = getattr(self, attr, None)
        return d

    def __repr__(self):
        return str(self.to_mapping())

    def __eq__(self, other):
        return self.to_mapping() == other.to_mapping()



from transiter import exceptions
from transiter.data import dbconnection
from transiter.data.dams import tripdam, routedam
from transiter.services import links, constants as c


@dbconnection.unit_of_work
def list_all_in_route(system_id, route_id, return_links=False):
    """
    Get representations for all trips in a system.

    :param system_id: the text id of the system
    :param route_id: the route id of the system
    :param return_links: whether to return links
    :return: a list of short model.Trip representations.
    """
    route = routedam.get_in_system_by_id(system_id, route_id)
    if route is None:
        raise exceptions.IdNotFoundError

    response = []
    trips = tripdam.list_all_in_route_by_pk(route.pk)
    trip_pk_to_last_stop = tripdam.get_trip_pk_to_last_stop_map(
        trip.pk for trip in trips
    )
    for trip in trips:
        last_stop = trip_pk_to_last_stop.get(trip.pk)
        trip_response = {**trip.to_dict(), c.LAST_STOP: last_stop.to_dict()}
        if return_links:
            trip_response[c.HREF] = links.TripEntityLink(trip)
            trip_response[c.LAST_STOP][c.HREF] = links.StopEntityLink(last_stop)
        response.append(trip_response)
    return response


@dbconnection.unit_of_work
def get_in_route_by_id(system_id, route_id, trip_id, return_links=False):
    """
    Get a representation for a trip in a system

    :param system_id: the text id of the system
    :param route_id: the text id of the route
    :param trip_id: the text id of the route
    :param return_links: whether to return links
    """
    trip = tripdam.get_in_route_by_id(system_id, route_id, trip_id)
    if trip is None:
        raise exceptions.IdNotFoundError
    trip_response = {
        **trip.to_large_dict(),
        c.ROUTE: trip.route.to_dict(),
        c.STOP_TIMES: [],
    }
    if return_links:
        trip_response[c.ROUTE][c.HREF] = links.RouteEntityLink(trip.route)
    for stop_time in trip.stop_times:
        stop_time_response = {
            **stop_time.to_dict(),
            c.STOP: stop_time.stop.to_dict(),
        }
        if return_links:
            stop_time_response[c.STOP][c.HREF] = links.StopEntityLink(stop_time.stop)
        trip_response[c.STOP_TIMES].append(stop_time_response)
    return trip_response



import datetime

from sqlalchemy import (
    Column,
    TIMESTAMP,
    DateTime,
    Index,
    Integer,
    String,
    Boolean,
    UniqueConstraint,
    ForeignKey,
)
from sqlalchemy.orm import relationship

from .base import Base, ToDictMixin


class TripStopTime(ToDictMixin, Base):
    __tablename__ = "trip_stop_time"

    pk = Column(Integer, primary_key=True)
    stop_pk = Column(Integer, ForeignKey("stop.pk"), nullable=False)
    trip_pk = Column(Integer, ForeignKey("trip.pk"), nullable=False)

    future = Column(Boolean, default=True, nullable=False)
    arrival_time = Column(TIMESTAMP(timezone=True))
    arrival_delay = Column(Integer)
    arrival_uncertainty = Column(Integer)
    departure_time = Column(DateTime(timezone=True))
    departure_delay = Column(Integer)
    departure_uncertainty = Column(Integer)
    stop_sequence = Column(Integer, nullable=False)
    track = Column(String)

    stop_id = None

    stop = relationship("Stop", back_populates="trip_times", cascade="none")
    trip = relationship(
        "Trip", back_populates="stop_times", cascade="none", cascade_backrefs=False
    )

    __table_args__ = (
        UniqueConstraint(trip_pk, stop_sequence),
        Index("trip_stop_time_stop_pk_arrival_time_idx", stop_pk, arrival_time),
    )

    _short_repr_list = [arrival_time, departure_time, track, future, stop_sequence]

    def to_dict(self):
        return {
            "arrival": {
                "time": self.arrival_time,
                "delay": self.arrival_delay,
                "uncertainty": self.arrival_uncertainty,
            },
            "departure": {
                "time": self.departure_time,
                "delay": self.departure_delay,
                "uncertainty": self.departure_uncertainty,
            },
            "track": self.track,
            "future": self.future,
            "stop_sequence": self.stop_sequence,
        }

    def get_time(self):
        if self.arrival_time is not None:
            return self.arrival_time
        return self.departure_time

    __mapping_columns__ = [
        pk,
        stop_pk,
        trip_pk,
        future,
        arrival_delay,
        arrival_time,
        arrival_uncertainty,
        departure_delay,
        departure_time,
        departure_uncertainty,
        stop_sequence,
        track,
    ]

    def to_mapping(self):
        return self._to_dict(self.__mapping_columns__)

    @staticmethod
    def from_feed(
        *,
        trip_id: str,
        stop_id: str,
        stop_sequence: int = None,
        future: bool = True,
        departure_time: datetime.datetime = None,
        departure_delay: int = None,
        departure_uncertainty: int = None,
        arrival_time: datetime.datetime = None,
        arrival_delay: int = None,
        arrival_uncertainty: int = None,
        track: str = None,
    ):
        return TripStopTimeLight(
            trip_id=trip_id,
            stop_id=stop_id,
            stop_sequence=stop_sequence,
            future=future,
            departure_time=departure_time,
            departure_delay=departure_delay,
            departure_uncertainty=departure_uncertainty,
            arrival_time=arrival_time,
            arrival_delay=arrival_delay,
            arrival_uncertainty=arrival_uncertainty,
            track=track,
        )


class TripStopTimeLight:

    __user_db_fields__ = {
        "future",
        "arrival_delay",
        "arrival_time",
        "arrival_uncertainty",
        "departure_delay",
        "departure_time",
        "departure_uncertainty",
        "stop_sequence",
        "track",
    }
    __user_non_db_fields = {
        "trip_id",
        "stop_id",
    }
    __other_db_fields__ = {
        "pk",
        "stop_pk",
        "trip_pk",
    }

    pk = None

    def __init__(
        self, **kwargs,
    ):
        for key, value in kwargs.items():
            if (
                key not in self.__user_db_fields__
                and key not in self.__user_non_db_fields
            ):
                raise KeyError(
                    "Unexpected argument to TripStopTime constructor: {}".format(key)
                )
            setattr(self, key, value)

    def to_mapping(self):
        d = {}
        for attr in self.__user_db_fields__:
            d[attr] = getattr(self, attr, None)
        for attr in self.__other_db_fields__:
            d[attr] = getattr(self, attr, None)
        return d

    def __repr__(self):
        return str(self.to_mapping())

    def __eq__(self, other):
        return self.to_mapping() == other.to_mapping()



import enum

from sqlalchemy import Column, Integer, String, ForeignKey, Enum, UniqueConstraint
from sqlalchemy.orm import relationship

from .base import Base, ToDictMixin
from .updatableentity import updatable_entity


@updatable_entity
class Route(ToDictMixin, Base):
    __tablename__ = "route"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    system_pk = Column(Integer, ForeignKey("system.pk"), nullable=False)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    class Type(enum.Enum):
        LIGHT_RAIL = 0
        SUBWAY = 1
        RAIL = 2
        BUS = 3
        FERRY = 4
        CABLE_CAR = 5
        GONDOLA = 6
        FUNICULAR = 7

    color = Column(String)
    text_color = Column(String)
    short_name = Column(String)
    long_name = Column(String)
    description = Column(String)
    url = Column(String)
    type = Column(Enum(Type, name="route_type"))
    sort_order = Column(Integer)

    system = relationship("System", back_populates="routes")
    source = relationship("FeedUpdate", cascade="none")
    trips = relationship("Trip", back_populates="route", cascade="all, delete-orphan")
    scheduled_trips = relationship(
        "ScheduledTrip", back_populates="route", cascade="all, delete-orphan"
    )
    service_maps = relationship(
        "ServiceMap",
        back_populates="route",
        primaryjoin="ServiceMap.route_pk==Route.pk",
        cascade="all, delete-orphan",
    )
    route_statuses = relationship(
        "Alert", secondary="alert_route", back_populates="routes", cascade="all"
    )

    __table_args__ = (UniqueConstraint("system_pk", "id"),)

    __dict_columns__ = [id, color]
    __large_dict_columns__ = [id, short_name, long_name, color, description, url, type]



__version__ = "0.4.2"

# The following are set by Travis CI at build time.
__build_number__ = None
__build_timestamp__ = None
__build_href__ = None
__git_commit_hash__ = None

# NOTE: the following were set by Travis CI
__version__ = "0.4.2"
__build_number__ = 738
__build_timestamp__ = 1587235180
__build_href__ = "https://travis-ci.org/jamespfennell/transiter/jobs/676666848"
__git_commit_hash__ = "ae629327d29759f0b4554cc2b4072ed52dfa9575"



from sqlalchemy import Column, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import relationship

from .base import Base
from .updatableentity import updatable_entity


@updatable_entity
class ScheduledService(Base):
    __tablename__ = "scheduled_service"

    pk = Column(Integer, primary_key=True)
    id = Column(String, nullable=False)
    system_pk = Column(Integer, ForeignKey("system.pk"), index=True)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    monday = Column(Boolean)
    tuesday = Column(Boolean)
    wednesday = Column(Boolean)
    thursday = Column(Boolean)
    friday = Column(Boolean)
    saturday = Column(Boolean)
    sunday = Column(Boolean)

    source = relationship("FeedUpdate", cascade="none")
    system = relationship("System", back_populates="scheduled_services")
    trips = relationship(
        "ScheduledTrip", back_populates="service", cascade="delete, delete-orphan"
    )



# This module provides a registry for updatable entities.


def updatable_entity(cls):
    _updatable_entities.append(cls)
    return cls


def list_updatable_entities():
    return _updatable_entities


_updatable_entities = []



"""
The feed service is used to retrieve data about feeds and perform feed update
operations.
"""

import datetime
import logging
from dataclasses import dataclass
from typing import Iterable

from transiter import exceptions
from transiter.data import dbconnection
from transiter.data.dams import feeddam, systemdam
from transiter.services import links, constants as c
from transiter.services.update import updatemanager

logger = logging.getLogger(__name__)


@dataclass
class Feed:
    system_id: str
    feed_id: str
    auto_update_period: int


@dbconnection.unit_of_work
def list_all_auto_updating() -> Iterable[Feed]:
    """
    List all auto updating feeds. This method is designed for use by the task
    server.
    """
    return [
        Feed(feed.system.id, feed.id, feed.auto_update_period)
        for feed in feeddam.list_all_auto_updating()
    ]


@dbconnection.unit_of_work
def list_all_in_system(system_id, return_links=True):
    """
    Get data on all feeds in a system.

    The result is list of dictionaries, one for each feed, containing the
    feed's short representation and (optionally) a link to the feed.

    :param system_id: the system ID
    :type system_id: str
    :param return_links: whether to return links
    :type return_links: bool
    :type system_id: str
    :return: the list described above
    :rtype: list
    """
    system = systemdam.get_by_id(system_id, only_return_active=True)
    if system is None:
        raise exceptions.IdNotFoundError
    response = []
    for feed in feeddam.list_all_in_system(system_id):
        feed_response = feed.to_dict()
        if return_links:
            feed_response[c.HREF] = links.FeedEntityLink(feed)
        response.append(feed_response)
    return response


@dbconnection.unit_of_work
def get_in_system_by_id(system_id, feed_id, return_links=True):
    """
    Get data on a specific feed in a system.

    :param system_id: the system ID
    :type system_id: str
    :param feed_id: the feed ID
    :type feed_id: str
    :param return_links: whether to return a link to the feed's updates page.
    :type return_links: bool
    :return: the feed's short representation.
    :rtype: dict
    """
    feed = feeddam.get_in_system_by_id(system_id, feed_id)
    if feed is None:
        raise exceptions.IdNotFoundError
    response = feed.to_dict()
    if return_links:
        response[c.UPDATES] = {c.HREF: links.FeedEntityUpdatesLink(feed)}
    return response


def create_and_execute_feed_update(
    system_id, feed_id, content=None, execute_async=False
):
    return _create_and_execute_feed_update_helper(
        system_id,
        feed_id,
        update_manager_function=updatemanager.create_feed_update,
        content=content,
        execute_async=execute_async,
    )


def create_and_execute_feed_flush(system_id, feed_id, execute_async=False):
    return _create_and_execute_feed_update_helper(
        system_id,
        feed_id,
        update_manager_function=updatemanager.create_feed_flush,
        content=None,
        execute_async=execute_async,
    )


def _create_and_execute_feed_update_helper(
    system_id, feed_id, update_manager_function, content=None, execute_async=False,
):
    """
    Create a feed update for a feed in a system.
    """
    feed_update_pk = update_manager_function(system_id, feed_id)
    if feed_update_pk is None:
        raise exceptions.IdNotFoundError
    if execute_async:
        updatemanager.execute_feed_update_async.delay(feed_update_pk, content)
    else:
        updatemanager.execute_feed_update(feed_update_pk, content)
    return feed_update_pk


@dbconnection.unit_of_work
def list_updates_in_feed(system_id, feed_id):
    """
    List all of the updates for a feed.

    :param system_id: the system ID
    :type system_id: str
    :param feed_id: the feed ID
    :type feed_id: str
    :return: a list of short representations of the feed updates.
    :rtype: list
    """
    feed = feeddam.get_in_system_by_id(system_id, feed_id)
    if feed is None:
        raise exceptions.IdNotFoundError
    response = []
    for feed_update in feeddam.list_updates_in_feed(feed.pk):
        response.append(feed_update.to_dict())
    return response


@dbconnection.unit_of_work
def get_update_in_feed_by_pk(system_id, feed_id, feed_update_pk):
    feed_update = feeddam.get_update_by_pk(feed_update_pk)
    if feed_update is None:
        raise exceptions.IdNotFoundError
    return feed_update.to_dict()


def trim_feed_updates():
    """
    Delete old feed updates.

    This method is designed to be called hourly by the task server.
    """

    @dbconnection.unit_of_work
    def _list_all_feed_pks():
        return feeddam.list_all_feed_pks()

    @dbconnection.unit_of_work
    def _trim_feed_updates_helper(feed_pk_, before_datetime_):
        logger.info(
            "Deleting feed updates with feed_pk={} and last updated before {}".format(
                feed_pk_, before_datetime_
            )
        )
        feeddam.trim_feed_updates(feed_pk_, before_datetime_)

    logger.info("Trimming old feed updates.")
    before_datetime = (
        datetime.datetime.now() - datetime.timedelta(minutes=60)
    ).replace(microsecond=0, second=0)

    for feed_pk in _list_all_feed_pks():
        _trim_feed_updates_helper(feed_pk, before_datetime)



import enum

from sqlalchemy import Column, TIMESTAMP, Table, Integer, String, ForeignKey, Enum
from sqlalchemy.orm import relationship

from .base import Base, ToDictMixin
from .updatableentity import updatable_entity


@updatable_entity
class Alert(ToDictMixin, Base):
    __tablename__ = "alert"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    class Cause(enum.Enum):
        UNKNOWN_CAUSE = 1
        OTHER_CAUSE = 2
        TECHNICAL_PROBLEM = 3
        STRIKE = 4
        DEMONSTRATION = 5
        ACCIDENT = 6
        HOLIDAY = 7
        WEATHER = 8
        MAINTENANCE = 9
        CONSTRUCTION = 10
        POLICE_ACTIVITY = 11
        MEDICAL_EMERGENCY = 12

    class Effect(enum.Enum):
        NO_SERVICE = 0
        REDUCED_SERVICE = 1
        SIGNIFICANT_DELAYS = 2
        DETOUR = 3
        ADDITIONAL_SERVICE = 4
        MODIFIED_SERVICE = 5
        OTHER_EFFECT = 6
        UNKNOWN_EFFECT = 7
        STOP_MOVED = 8

    header = Column(String)
    description = Column(String)
    url = Column(String)

    cause = Column(Enum(Cause))
    effect = Column(Enum(Effect))
    priority = Column(Integer)
    start_time = Column(TIMESTAMP(timezone=True))
    end_time = Column(TIMESTAMP(timezone=True))
    creation_time = Column(TIMESTAMP(timezone=True))

    # NOTE: the following foreign key is *temporary*! It is used when, in a feed, the
    # alert's selector references an agency. Pending the introduction of a
    # models.Agency type with which we can many-to-many relate, we just relate such an
    # alert to the system it is in.
    system_pk = Column(Integer, ForeignKey("system.pk"), index=True)

    route_ids = set()
    agency_ids = set()

    source = relationship("FeedUpdate", cascade="none")
    routes = relationship(
        "Route",
        secondary="alert_route",
        back_populates="route_statuses",
        cascade="none",
    )

    __dict_columns__ = [id, start_time, end_time, creation_time, header, description]
    __large_dict_columns__ = [
        id,
        start_time,
        end_time,
        creation_time,
        header,
        description,
        url,
        cause,
        effect,
    ]


alert_route = Table(
    "alert_route",
    Base.metadata,
    Column("alert_pk", Integer, ForeignKey("alert.pk")),
    Column("route_pk", Integer, ForeignKey("route.pk"), index=True),
)



from sqlalchemy import Index, Column, Integer, String, Boolean, ForeignKey
from sqlalchemy.orm import relationship

from .base import Base
from .updatableentity import updatable_entity


@updatable_entity
class DirectionRule(Base):
    __tablename__ = "direction_name_rule"

    pk = Column(Integer, primary_key=True)
    id = Column(String)
    stop_pk = Column(Integer, ForeignKey("stop.pk"))
    source_pk = Column(Integer, ForeignKey("feed_update.pk"), index=True)

    priority = Column(Integer)
    direction_id = Column(Boolean)
    track = Column(String)
    name = Column(String)

    stop_id = None

    source = relationship("FeedUpdate", cascade="none")
    stop = relationship("Stop", back_populates="direction_rules", cascade="none")

    __table_args__ = (
        Index("direction_name_rule_stop_pk_priority_idx", stop_pk, priority),
    )



import flask
import requests

from transiter import exceptions
from transiter.http import httpmanager
from transiter.http.httpmanager import (
    http_endpoint,
    link_target,
    HttpMethod,
    HttpStatus,
)
from transiter.http.permissions import requires_permissions, PermissionsLevel
from transiter.services import systemservice, links

system_endpoints = flask.Blueprint(__name__, __name__)


@http_endpoint(system_endpoints, "")
@link_target(links.SystemsIndexLink)
def list_all():
    """List all systems"""
    return systemservice.list_all()


@http_endpoint(system_endpoints, "/<system_id>")
@link_target(links.SystemEntityLink)
def get_by_id(system_id):
    """Get data on a specific system."""
    return systemservice.get_by_id(system_id)


@http_endpoint(
    system_endpoints, "/<system_id>", method=HttpMethod.PUT,
)
@requires_permissions(PermissionsLevel.ALL)
def install(system_id):
    """Install a system."""
    form_key_to_value = flask.request.form.to_dict()
    form_key_to_file_storage = flask.request.files.to_dict()

    config_source_url = None
    if "config_file" in form_key_to_value:
        config_source_url = form_key_to_value["config_file"]
        try:
            response = requests.get(config_source_url)
            response.raise_for_status()
        except requests.exceptions.RequestException:
            raise exceptions.InvalidInput(
                "Could not download YAML config file from '{}'".format(
                    config_source_url
                )
            )
        config_str = response.text
        del form_key_to_value["config_file"]
    elif "config_file" in form_key_to_file_storage:
        config_str = flask.request.files["config_file"].read().decode("utf-8")
    else:
        raise exceptions.InvalidInput("YAML config file not provided!")

    # TODO: replace with httpmanager.get_boolean_url_parameter("sync")
    sync = httpmanager.is_sync_request()
    response = systemservice.install(
        system_id=system_id,
        config_str=config_str,
        extra_settings=form_key_to_value,
        config_source_url=config_source_url,
        sync=sync,
    )

    # This means the system already exists and nothing was done.
    # TODO: this is all wrong
    if not response:
        status = HttpStatus.OK
    else:
        if sync:
            status = HttpStatus.CREATED
        else:
            status = HttpStatus.ACCEPTED
    # TODO return a representation of the update instead
    return systemservice.get_by_id(system_id), status


# TODO: use this method
def _get_config_file(config_source_url, uploaded_config_file):
    if config_source_url is not None:
        try:
            response = requests.get(config_source_url)
            response.raise_for_status()
        except requests.exceptions.RequestException:
            raise exceptions.InvalidInput(
                "Could not download YAML config file from '{}'".format(
                    config_source_url
                )
            )
        return response.text
    elif uploaded_config_file is not None:
        return uploaded_config_file.read().decode("utf-8")
    else:
        raise exceptions.InvalidInput("YAML config file not provided!")


@http_endpoint(
    system_endpoints,
    "/<system_id>",
    method=HttpMethod.DELETE,
    returns_json_response=False,
)
@requires_permissions(PermissionsLevel.ALL)
def delete_by_id(system_id):
    """Uninstall a system."""
    systemservice.delete_by_id(
        system_id, error_if_not_exists=True, sync=httpmanager.is_sync_request()
    )
    return flask.Response(response="", status=HttpStatus.NO_CONTENT, content_type="")


@http_endpoint(
    system_endpoints, "/<system_id>/auto-update", method=HttpMethod.PUT,
)
@requires_permissions(PermissionsLevel.ALL)
def set_auto_update_enabled(system_id):
    form_key_to_value = flask.request.form.to_dict()
    enabled = form_key_to_value.get("enabled")
    if enabled is None:
        raise exceptions.InvalidInput("The form variable 'enabled' is required")
    enabled = enabled.lower()
    if enabled not in {"false", "true"}:
        raise exceptions.InvalidInput(
            "The form variable 'enabled' has to be 'true' or 'false', not '{}'".format(
                enabled
            )
        )
    systemservice.set_auto_update_enabled(
        system_id, form_key_to_value["enabled"].lower() == "true"
    )
    return "", HttpStatus.NO_CONTENT
